use strict;
use warnings;

# this test was generated with Dist::Zilla::Plugin::Test::NoTabs 0.15

use Test::More 0.88;
use Test::NoTabs;

my @files = (
    'lib/File/KDBX.pm',
    'lib/File/KDBX/Cipher.pm',
    'lib/File/KDBX/Cipher/CBC.pm',
    'lib/File/KDBX/Cipher/Stream.pm',
    'lib/File/KDBX/Constants.pm',
    'lib/File/KDBX/Dumper.pm',
    'lib/File/KDBX/Dumper/KDB.pm',
    'lib/File/KDBX/Dumper/Raw.pm',
    'lib/File/KDBX/Dumper/V3.pm',
    'lib/File/KDBX/Dumper/V4.pm',
    'lib/File/KDBX/Dumper/XML.pm',
    'lib/File/KDBX/Entry.pm',
    'lib/File/KDBX/Error.pm',
    'lib/File/KDBX/Group.pm',
    'lib/File/KDBX/IO.pm',
    'lib/File/KDBX/IO/Crypt.pm',
    'lib/File/KDBX/IO/HashBlock.pm',
    'lib/File/KDBX/IO/HmacBlock.pm',
    'lib/File/KDBX/Iterator.pm',
    'lib/File/KDBX/KDF.pm',
    'lib/File/KDBX/KDF/AES.pm',
    'lib/File/KDBX/KDF/Argon2.pm',
    'lib/File/KDBX/Key.pm',
    'lib/File/KDBX/Key/ChallengeResponse.pm',
    'lib/File/KDBX/Key/Composite.pm',
    'lib/File/KDBX/Key/File.pm',
    'lib/File/KDBX/Key/Password.pm',
    'lib/File/KDBX/Key/YubiKey.pm',
    'lib/File/KDBX/Loader.pm',
    'lib/File/KDBX/Loader/KDB.pm',
    'lib/File/KDBX/Loader/Raw.pm',
    'lib/File/KDBX/Loader/V3.pm',
    'lib/File/KDBX/Loader/V4.pm',
    'lib/File/KDBX/Loader/XML.pm',
    'lib/File/KDBX/Object.pm',
    'lib/File/KDBX/Safe.pm',
    'lib/File/KDBX/Transaction.pm',
    'lib/File/KDBX/Util.pm',
    't/00-compile.t',
    't/00-report-prereqs.dd',
    't/00-report-prereqs.t',
    't/crypt.t',
    't/database.t',
    't/entry.t',
    't/erase.t',
    't/error.t',
    't/files/bin/ykchalresp',
    't/files/bin/ykinfo',
    't/group.t',
    't/hash-block.t',
    't/hmac-block.t',
    't/iterator.t',
    't/kdb.t',
    't/kdbx2.t',
    't/kdbx3.t',
    't/kdbx4.t',
    't/kdf-aes-pp.t',
    't/kdf.t',
    't/keys.t',
    't/lib/TestCommon.pm',
    't/memory-protection.t',
    't/object.t',
    't/otp.t',
    't/placeholders.t',
    't/query.t',
    't/references.t',
    't/safe.t',
    't/util.t',
    't/yubikey.t',
    'xt/author/clean-namespaces.t',
    'xt/author/critic.t',
    'xt/author/distmeta.t',
    'xt/author/eol.t',
    'xt/author/minimum-version.t',
    'xt/author/no-tabs.t',
    'xt/author/pod-coverage.t',
    'xt/author/pod-no404s.t',
    'xt/author/pod-syntax.t',
    'xt/author/portability.t',
    'xt/release/cpan-changes.t'
);

notabs_ok($_) foreach @files;
done_testing;
